"""RFSoC FPGA driver."""

from dataclasses import asdict, dataclass
from typing import Union

import numpy as np
import numpy.typing as npt
import qibosoq.components.base as rfsoc
from qibosoq import client

from qibolab import AcquisitionType, AveragingMode, ExecutionParameters
from qibolab.instruments.abstract import Controller
from qibolab.instruments.port import Port
from qibolab.platform import Qubit
from qibolab.pulses import PulseSequence, PulseType
from qibolab.result import IntegratedResults, SampleResults
from qibolab.sweeper import BIAS, Sweeper

from .convert import convert, convert_units_sweeper

HZ_TO_MHZ = 1e-6
NS_TO_US = 1e-3


@dataclass
class RFSoCPort(Port):
    """Port object of the RFSoC."""

    name: int
    """DAC number."""
    offset: float = 0.0
    """Amplitude factor for biasing."""


class RFSoC(Controller):
    """Instrument object for controlling RFSoC FPGAs.

    The two way of executing pulses are with ``play`` (for arbitrary
    qibolab ``PulseSequence``) or with ``sweep`` that execute a
    ``PulseSequence`` object with one or more ``Sweeper``.

    Attributes:
        cfg (rfsoc.Config): Configuration dictionary required for pulse execution.
    """

    PortType = RFSoCPort

    def __init__(self, name: str, address: str, port: int):
        """Set server information and base configuration.

        Args:
            name (str): Name of the instrument instance.
            address (str): IP and port of the server (ex. 192.168.0.10)
            port (int): Port of the server (ex.6000)
        """
        super().__init__(name, address=address)
        self.host = address
        self.port = port
        self.cfg = rfsoc.Config()

    def connect(self):
        """Empty method to comply with Instrument interface."""

    def start(self):
        """Empty method to comply with Instrument interface."""

    def stop(self):
        """Empty method to comply with Instrument interface."""

    def disconnect(self):
        """Empty method to comply with Instrument interface."""

    def setup(self):
        """Empty deprecated method."""

    def _execute_pulse_sequence(
        self, sequence: PulseSequence, qubits: dict[int, Qubit], opcode: rfsoc.OperationCode
    ) -> tuple[list, list]:
        """Prepare the commands dictionary to send to the qibosoq server.

        Args:
            sequence (`qibolab.pulses.PulseSequence`): arbitrary PulseSequence object to execute
            qubits: list of qubits (`qibolab.platforms.abstract.Qubit`) of the platform in the form of a dictionary
            opcode: can be `rfsoc.OperationCode.EXECUTE_PULSE_SEQUENCE` or `rfsoc.OperationCode.EXECUTE_PULSE_SEQUENCE_RAW`
        Returns:
            Lists of I and Q value measured
        """
        server_commands = {
            "operation_code": opcode,
            "cfg": asdict(self.cfg),
            "sequence": convert(sequence, qubits),
            "qubits": [asdict(convert(qubits[idx])) for idx in qubits],
        }
        return client.connect(server_commands, self.host, self.port)

    def _execute_sweeps(
        self,
        sequence: PulseSequence,
        qubits: dict[int, Qubit],
        sweepers: list[rfsoc.Sweeper],
    ) -> tuple[list, list]:
        """Prepare the commands dictionary to send to the qibosoq server.

        Args:
            sequence (`qibolab.pulses.PulseSequence`): arbitrary PulseSequence object to execute
            qubits: list of qubits (`qibolab.platforms.abstract.Qubit`) of the platform in the form of a dictionary
            sweepers: list of `qibosoq.abstract.Sweeper` objects
        Returns:
            Lists of I and Q value measured
        """
        for sweeper in sweepers:
            convert_units_sweeper(sweeper, sequence, qubits)
        server_commands = {
            "operation_code": rfsoc.OperationCode.EXECUTE_SWEEPS,
            "cfg": asdict(self.cfg),
            "sequence": convert(sequence, qubits),
            "qubits": [asdict(convert(qubits[idx])) for idx in qubits],
            "sweepers": [sweeper.serialized for sweeper in sweepers],
        }
        return client.connect(server_commands, self.host, self.port)

    def play(
        self,
        qubits: dict[int, Qubit],
        sequence: PulseSequence,
        execution_parameters: ExecutionParameters,
    ) -> dict[str, Union[IntegratedResults, SampleResults]]:
        """Execute the sequence of instructions and retrieves readout results.

        Each readout pulse generates a separate acquisition.
        The relaxation_time and the number of shots have default values.

        Args:
            qubits (dict): List of `qibolab.platforms.utils.Qubit` objects
                           passed from the platform.
            execution_parameters (`qibolab.ExecutionParameters`): Parameters (nshots,
                                                        relaxation_time,
                                                        fast_reset,
                                                        acquisition_type,
                                                        averaging_mode)
            sequence (`qibolab.pulses.PulseSequence`): Pulse sequence to play.
        Returns:
            A dictionary mapping the readout pulses serial and respective qubits to
            qibolab results objects
        """
        self.validate_input_command(sequence, execution_parameters, sweep=False)
        self.update_cfg(execution_parameters)

        if execution_parameters.acquisition_type is AcquisitionType.DISCRIMINATION:
            self.cfg.average = False
        else:
            self.cfg.average = execution_parameters.averaging_mode is AveragingMode.CYCLIC

        if execution_parameters.acquisition_type is AcquisitionType.RAW:
            opcode = rfsoc.OperationCode.EXECUTE_PULSE_SEQUENCE_RAW
        else:
            opcode = rfsoc.OperationCode.EXECUTE_PULSE_SEQUENCE
        toti, totq = self._execute_pulse_sequence(sequence, qubits, opcode)

        results = {}
        probed_qubits = np.unique([p.qubit for p in sequence.ro_pulses])

        for j, qubit in enumerate(probed_qubits):
            for i, ro_pulse in enumerate(sequence.ro_pulses.get_qubit_pulses(qubit)):
                i_pulse = np.array(toti[j][i])
                q_pulse = np.array(totq[j][i])

                if execution_parameters.acquisition_type is AcquisitionType.DISCRIMINATION:
                    discriminated_shots = self.classify_shots(i_pulse, q_pulse, qubits[ro_pulse.qubit])
                    if execution_parameters.averaging_mode is AveragingMode.CYCLIC:
                        discriminated_shots = np.mean(discriminated_shots, axis=0)
                    result = execution_parameters.results_type(discriminated_shots)
                else:
                    result = execution_parameters.results_type(i_pulse + 1j * q_pulse)
                results[ro_pulse.qubit] = results[ro_pulse.serial] = result

        return results

    def play_sequences(self, qubits, sequence, options):
        raise NotImplementedError

    @staticmethod
    def validate_input_command(sequence: PulseSequence, execution_parameters: ExecutionParameters, sweep: bool):
        """Check if sequence and execution_parameters are supported."""
        if execution_parameters.acquisition_type is AcquisitionType.RAW:
            if sweep:
                raise NotImplementedError("Raw data acquisition is not compatible with sweepers")
            if len(sequence.ro_pulses) != 1:
                raise NotImplementedError("Raw data acquisition is compatible only with a single readout")
            if execution_parameters.averaging_mode is not AveragingMode.CYCLIC:
                raise NotImplementedError("Raw data acquisition can only be averaged")
        if execution_parameters.fast_reset:
            raise NotImplementedError("Fast reset is not supported")

    def update_cfg(self, execution_parameters: ExecutionParameters):
        """Update rfsoc.Config object with new parameters."""
        if execution_parameters.nshots is not None:
            self.cfg.reps = execution_parameters.nshots
        if execution_parameters.relaxation_time is not None:
            self.cfg.repetition_duration = execution_parameters.relaxation_time * NS_TO_US

    def classify_shots(
        self, i_values: npt.NDArray[np.float64], q_values: npt.NDArray[np.float64], qubit: Qubit
    ) -> npt.NDArray[np.float64]:
        """Classify IQ values using qubit threshold and rotation_angle if available in runcard."""
        if qubit.iq_angle is None or qubit.threshold is None:
            raise ValueError("Classification parameters were not provided")
        angle = qubit.iq_angle
        threshold = qubit.threshold

        rotated = np.cos(angle) * np.array(i_values) - np.sin(angle) * np.array(q_values)
        shots = np.heaviside(np.array(rotated) - threshold, 0)
        if isinstance(shots, float):
            return [shots]
        return shots

    def play_sequence_in_sweep_recursion(
        self,
        qubits: dict[int, Qubit],
        sequence: PulseSequence,
        or_sequence: PulseSequence,
        execution_parameters: ExecutionParameters,
    ) -> dict[str, Union[IntegratedResults, SampleResults]]:
        """Last recursion layer, if no sweeps are present.

        After playing the sequence, the resulting dictionary keys need
        to be converted to the correct values.
        Even indexes correspond to qubit number and are not changed.
        Odd indexes correspond to readout pulses serials and are convert
        to match the original sequence (of the sweep) and not the one just executed.
        """
        res = self.play(qubits, sequence, execution_parameters)
        newres = {}
        serials = [pulse.serial for pulse in or_sequence.ro_pulses]
        for idx, key in enumerate(res):
            if idx % 2 == 1:
                newres[serials[idx // 2]] = res[key]
            else:
                newres[key] = res[key]

        return newres

    def recursive_python_sweep(
        self,
        qubits: dict[int, Qubit],
        sequence: PulseSequence,
        or_sequence: PulseSequence,
        *sweepers: rfsoc.Sweeper,
        execution_parameters: ExecutionParameters,
    ) -> dict[str, Union[IntegratedResults, SampleResults]]:
        """Execute a sweep of an arbitrary number of Sweepers via recursion.

        Args:
            qubits (list): List of `qibolab.platforms.utils.Qubit` objects
                    passed from the platform.
            sequence (`qibolab.pulses.PulseSequence`): Pulse sequence to play.
                    This object is a deep copy of the original
                    sequence and gets modified.
            or_sequence (`qibolab.pulses.PulseSequence`): Reference to original
                    sequence to not modify.
            *sweepers (`qibolab.Sweeper`): Sweeper objects.
            execution_parameters (`qibolab.ExecutionParameters`): Parameters (nshots,
                                                        relaxation_time,
                                                        fast_reset,
                                                        acquisition_type,
                                                        averaging_mode)
        Returns:
            A dictionary mapping the readout pulses serial and respective qubits to
            results objects
        """
        # If there are no sweepers run ExecutePulseSequence acquisition.
        # Last layer for recursion.

        if len(sweepers) == 0:
            return self.play_sequence_in_sweep_recursion(qubits, sequence, or_sequence, execution_parameters)

        if not self.get_if_python_sweep(sequence, *sweepers):
            toti, totq = self._execute_sweeps(sequence, qubits, sweepers)
            res = self.convert_sweep_results(or_sequence, qubits, toti, totq, execution_parameters)
            return res

        sweeper = sweepers[0]
        values = []
        for idx, _ in enumerate(sweeper.indexes):
            val = np.linspace(sweeper.starts[idx], sweeper.stops[idx], sweeper.expts)
            if sweeper.parameters[idx] in rfsoc.Parameter.variants({"duration", "delay"}):
                val = val.astype(int)
            values.append(val)

        results: dict[str, Union[IntegratedResults, SampleResults]] = {}
        for idx in range(sweeper.expts):
            # update values
            for jdx, kdx in enumerate(sweeper.indexes):
                sweeper_parameter = sweeper.parameters[jdx]
                if sweeper_parameter is rfsoc.Parameter.BIAS:
                    qubits[list(qubits)[kdx]].flux.offset = values[jdx][idx]
                elif sweeper_parameter in rfsoc.Parameter.variants(
                    {
                        "amplitude",
                        "frequency",
                        "relative_phase",
                        "duration",
                    }
                ):
                    setattr(sequence[kdx], sweeper_parameter.name.lower(), values[jdx][idx])
                elif sweeper is rfsoc.Parameter.DELAY:
                    start_delay = values[jdx][idx]
                    sequence[kdx].start_delay = values[jdx][idx]

            res = self.recursive_python_sweep(
                qubits, sequence, or_sequence, *sweepers[1:], execution_parameters=execution_parameters
            )
            results = self.merge_sweep_results(results, res)
        return results  # already in the right format

    @staticmethod
    def merge_sweep_results(
        dict_a: dict[str, Union[IntegratedResults, SampleResults]],
        dict_b: dict[str, Union[IntegratedResults, SampleResults]],
    ) -> dict[str, Union[IntegratedResults, SampleResults]]:
        """Merge two dictionary mapping pulse serial to Results object.

        If dict_b has a key (serial) that dict_a does not have, simply add it,
        otherwise sum the two results

        Args:
            dict_a (dict): dict mapping ro pulses serial to qibolab res objects
            dict_b (dict): dict mapping ro pulses serial to qibolab res objects
        Returns:
            A dict mapping the readout pulses serial to qibolab results objects
        """
        for serial in dict_b:
            if serial in dict_a:
                dict_a[serial] = dict_a[serial] + dict_b[serial]
            else:
                dict_a[serial] = dict_b[serial]
        return dict_a

    def get_if_python_sweep(self, sequence: PulseSequence, *sweepers: rfsoc.Sweeper) -> bool:
        """Check if a sweeper must be run with python loop or on hardware.

        To be run on qick internal loop a sweep must:
            * not be on the readout frequency
            * not be a duration sweeper
            * only one pulse per channel supported
            * flux pulses are not compatible with sweepers

        Args:
            sequence (`qibolab.pulses.PulseSequence`). Pulse sequence to play.
            *sweepers (`qibosoq.abstract.Sweeper`): Sweeper objects.
        Returns:
            A boolean value true if the sweeper must be executed by python
            loop, false otherwise
        """
        if any(pulse.type is PulseType.FLUX for pulse in sequence):
            return True
        for sweeper in sweepers:
            if all(parameter is rfsoc.Parameter.BIAS for parameter in sweeper.parameters):
                continue
            if all(parameter is rfsoc.Parameter.DELAY for parameter in sweeper.parameters):
                continue
            if any(parameter is rfsoc.Parameter.DURATION for parameter in sweeper.parameters):
                return True

            for sweep_idx, parameter in enumerate(sweeper.parameters):
                is_freq = parameter is rfsoc.Parameter.FREQUENCY
                is_ro = sequence[sweeper.indexes[sweep_idx]].type == PulseType.READOUT
                # if it's a sweep on the readout freq do a python sweep
                if is_freq and is_ro:
                    return True

            for idx in sweeper.indexes:
                sweep_pulse = sequence[idx]
                channel = sweep_pulse.channel
                ch_pulses = sequence.get_channel_pulses(channel)
                if len(ch_pulses) > 1:
                    return True
        # if all passed, do a firmware sweep
        return False

    def convert_sweep_results(
        self,
        original_ro: PulseSequence,
        qubits: dict[int, Qubit],
        toti: list[list[list[float]]],
        totq: list[list[list[float]]],
        execution_parameters: ExecutionParameters,
    ) -> dict[str, Union[IntegratedResults, SampleResults]]:
        """Convert sweep res to qibolab dict res.

        Args:
            original_ro (`qibolab.pulses.PulseSequence`): Original PulseSequence
            qubits (list): List of `qibolab.platforms.utils.Qubit` objects
                 passed from the platform.
            toti (list): i values
            totq (list): q values
            results_type: qibolab results object
            execution_parameters (`qibolab.ExecutionParameters`): Parameters (nshots,
                                                        relaxation_time,
                                                        fast_reset,
                                                        acquisition_type,
                                                        averaging_mode)
        Returns:
            A dict mapping the readout pulses serial to qibolab results objects
        """
        results = {}

        adcs = np.unique([qubits[p.qubit].feedback.port.name for p in original_ro])
        for k, k_val in enumerate(adcs):
            adc_ro = [pulse for pulse in original_ro if qubits[pulse.qubit].feedback.port.name == k_val]
            for i, (ro_pulse, original_ro_pulse) in enumerate(zip(adc_ro, original_ro)):
                i_vals = np.array(toti[k][i])
                q_vals = np.array(totq[k][i])

                if not self.cfg.average:
                    i_vals = np.reshape(i_vals, (self.cfg.reps, *i_vals.shape[:-1]))
                    q_vals = np.reshape(q_vals, (self.cfg.reps, *q_vals.shape[:-1]))

                if execution_parameters.acquisition_type is AcquisitionType.DISCRIMINATION:
                    qubit = qubits[original_ro_pulse.qubit]
                    discriminated_shots = self.classify_shots(i_vals, q_vals, qubit)
                    if execution_parameters.averaging_mode is AveragingMode.CYCLIC:
                        discriminated_shots = np.mean(discriminated_shots, axis=0)
                    result = execution_parameters.results_type(discriminated_shots)
                else:
                    result = execution_parameters.results_type(i_vals + 1j * q_vals)

                results[original_ro_pulse.qubit] = results[ro_pulse.serial] = result
        return results

    def sweep(
        self,
        qubits: dict[int, Qubit],
        sequence: PulseSequence,
        execution_parameters: ExecutionParameters,
        *sweepers: Sweeper,
    ) -> dict[str, Union[IntegratedResults, SampleResults]]:
        """Execute the sweep and retrieves the readout results.

        Each readout pulse generates a separate acquisition.
        The relaxation_time and the number of shots have default values.

        Args:
            qubits (list): List of `qibolab.platforms.utils.Qubit` objects
                           passed from the platform.
            execution_parameters (`qibolab.ExecutionParameters`): Parameters (nshots,
                                                        relaxation_time,
                                                        fast_reset,
                                                        acquisition_type,
                                                        averaging_mode)
            sequence (`qibolab.pulses.PulseSequence`). Pulse sequence to play.
            *sweepers (`qibolab.Sweeper`): Sweeper objects.
        Returns:
            A dictionary mapping the readout pulses serial and respective qubits to
            results objects
        """
        self.validate_input_command(sequence, execution_parameters, sweep=True)
        self.update_cfg(execution_parameters)

        if execution_parameters.acquisition_type is AcquisitionType.DISCRIMINATION:
            self.cfg.average = False
        else:
            self.cfg.average = execution_parameters.averaging_mode is AveragingMode.CYCLIC

        rfsoc_sweepers = [convert(sweep, sequence, qubits) for sweep in sweepers]

        sweepsequence = sequence.copy()

        bias_change = any(sweep.parameter is BIAS for sweep in sweepers)
        if bias_change:
            initial_biases = [qubits[idx].flux.offset if qubits[idx].flux is not None else None for idx in qubits]

        results = self.recursive_python_sweep(
            qubits,
            sweepsequence,
            sequence.ro_pulses,
            *rfsoc_sweepers,
            execution_parameters=execution_parameters,
        )

        if bias_change:
            for idx, qubit in enumerate(qubits.values()):
                if qubit.flux is not None:
                    qubit.flux.offset = initial_biases[idx]

        return results



import math

import numpy as np
from qibo.config import raise_error
from qm import qua
from qm.qua import declare, fixed, for_
from qualang_tools.loops import from_array

from qibolab.channels import check_max_offset
from qibolab.instruments.qm.sequence import BakedPulse
from qibolab.pulses import PulseType
from qibolab.sweeper import Parameter


def maximum_sweep_value(values, value0):
    """Calculates maximum value that is reached during a sweep.

    Useful to check whether a sweep exceeds the range of allowed values.
    Note that both the array of values we sweep and the center value can
    be negative, so we need to make sure that the maximum absolute value
    is within range.

    Args:
        values (np.ndarray): Array of values we will sweep over.
        value0 (float, int): Center value of the sweep.
    """
    return max(abs(min(values) + value0), abs(max(values) + value0))


def _update_baked_pulses(sweeper, qmsequence, config):
    """Updates baked pulse if duration sweeper is used."""
    qmpulse = qmsequence.pulse_to_qmpulse[sweeper.pulses[0].serial]
    is_baked = isinstance(qmpulse, BakedPulse)
    for pulse in sweeper.pulses:
        qmpulse = qmsequence.pulse_to_qmpulse[pulse.serial]
        if isinstance(qmpulse, BakedPulse):
            if not is_baked:
                raise_error(TypeError, "Duration sweeper cannot contain both baked and not baked pulses.")
            values = np.array(sweeper.values).astype(int)
            qmpulse.bake(config, values)


def sweep(sweepers, qubits, qmsequence, relaxation_time, config):
    """Public sweep function that is called by the driver."""
    for sweeper in sweepers:
        if sweeper.parameter is Parameter.duration:
            _update_baked_pulses(sweeper, qmsequence, config)
    _sweep_recursion(sweepers, qubits, qmsequence, relaxation_time)


def _sweep_recursion(sweepers, qubits, qmsequence, relaxation_time):
    """Unrolls a list of qibolab sweepers to the corresponding QUA for loops using recursion."""
    if len(sweepers) > 0:
        parameter = sweepers[0].parameter.name
        func_name = f"_sweep_{parameter}"
        if func_name in globals():
            globals()[func_name](sweepers, qubits, qmsequence, relaxation_time)
        else:
            raise_error(NotImplementedError, f"Sweeper for {parameter} is not implemented.")
    else:
        qmsequence.play(relaxation_time)


def _sweep_frequency(sweepers, qubits, qmsequence, relaxation_time):
    sweeper = sweepers[0]
    freqs0 = []
    for pulse in sweeper.pulses:
        qubit = qubits[pulse.qubit]
        if pulse.type is PulseType.DRIVE:
            lo_frequency = math.floor(qubit.drive.local_oscillator.frequency)
        elif pulse.type is PulseType.READOUT:
            lo_frequency = math.floor(qubit.readout.local_oscillator.frequency)
        else:
            raise_error(NotImplementedError, f"Cannot sweep frequency of pulse of type {pulse.type}.")
        # convert to IF frequency for readout and drive pulses
        f0 = math.floor(pulse.frequency - lo_frequency)
        freqs0.append(declare(int, value=f0))
        # check if sweep is within the supported bandwidth [-400, 400] MHz
        max_freq = maximum_sweep_value(sweeper.values, f0)
        if max_freq > 4e8:
            raise_error(ValueError, f"Frequency {max_freq} for qubit {qubit.name} is beyond instrument bandwidth.")

    # is it fine to have this declaration inside the ``nshots`` QUA loop?
    f = declare(int)
    with for_(*from_array(f, sweeper.values.astype(int))):
        for pulse, f0 in zip(sweeper.pulses, freqs0):
            qmpulse = qmsequence.pulse_to_qmpulse[pulse.serial]
            qua.update_frequency(qmpulse.element, f + f0)

        _sweep_recursion(sweepers[1:], qubits, qmsequence, relaxation_time)


def _sweep_amplitude(sweepers, qubits, qmsequence, relaxation_time):
    sweeper = sweepers[0]
    # TODO: Consider sweeping amplitude without multiplication
    if min(sweeper.values) < -2:
        raise_error(ValueError, "Amplitude sweep values are <-2 which is not supported.")
    if max(sweeper.values) > 2:
        raise_error(ValueError, "Amplitude sweep values are >2 which is not supported.")

    a = declare(fixed)
    with for_(*from_array(a, sweeper.values)):
        for pulse in sweeper.pulses:
            qmpulse = qmsequence.pulse_to_qmpulse[pulse.serial]
            if isinstance(qmpulse, BakedPulse):
                qmpulse.amplitude = a
            else:
                qmpulse.operation = qmpulse.operation * qua.amp(a)

        _sweep_recursion(sweepers[1:], qubits, qmsequence, relaxation_time)


def _sweep_relative_phase(sweepers, qubits, qmsequence, relaxation_time):
    sweeper = sweepers[0]
    relphase = declare(fixed)
    with for_(*from_array(relphase, sweeper.values / (2 * np.pi))):
        for pulse in sweeper.pulses:
            qmpulse = qmsequence.pulse_to_qmpulse[pulse.serial]
            qmpulse.relative_phase = relphase

        _sweep_recursion(sweepers[1:], qubits, qmsequence, relaxation_time)


def _sweep_bias(sweepers, qubits, qmsequence, relaxation_time):
    sweeper = sweepers[0]
    offset0 = []
    for qubit in sweeper.qubits:
        b0 = qubit.flux.offset
        max_offset = qubit.flux.max_offset
        max_value = maximum_sweep_value(sweeper.values, b0)
        check_max_offset(max_value, max_offset)
        offset0.append(declare(fixed, value=b0))
    b = declare(fixed)
    with for_(*from_array(b, sweeper.values)):
        for qubit, b0 in zip(sweeper.qubits, offset0):
            with qua.if_((b + b0) >= 0.49):
                qua.set_dc_offset(f"flux{qubit.name}", "single", 0.49)
            with qua.elif_((b + b0) <= -0.49):
                qua.set_dc_offset(f"flux{qubit.name}", "single", -0.49)
            with qua.else_():
                qua.set_dc_offset(f"flux{qubit.name}", "single", (b + b0))

        _sweep_recursion(sweepers[1:], qubits, qmsequence, relaxation_time)


def _sweep_start(sweepers, qubits, qmsequence, relaxation_time):
    sweeper = sweepers[0]
    start = declare(int)
    values = (np.array(sweeper.values) // 4).astype(int)

    if len(np.unique(values[1:] - values[:-1])) > 1:
        loop = qua.for_each_(start, values)
    else:
        loop = for_(*from_array(start, values))

    with loop:
        for pulse in sweeper.pulses:
            qmpulse = qmsequence.pulse_to_qmpulse[pulse.serial]
            # find all pulses that are connected to ``qmpulse`` and update their starts
            to_process = {qmpulse}
            while to_process:
                next_qmpulse = to_process.pop()
                to_process |= next_qmpulse.next_
                next_qmpulse.wait_time_variable = start

        _sweep_recursion(sweepers[1:], qubits, qmsequence, relaxation_time)


def _sweep_duration(sweepers, qubits, qmsequence, relaxation_time):
    sweeper = sweepers[0]
    qmpulse = qmsequence.pulse_to_qmpulse[sweeper.pulses[0].serial]
    if isinstance(qmpulse, BakedPulse):
        values = np.array(sweeper.values).astype(int)
    else:
        values = np.array(sweeper.values).astype(int) // 4

    dur = declare(int)
    with for_(*from_array(dur, values)):
        for pulse in sweeper.pulses:
            qmpulse = qmsequence.pulse_to_qmpulse[pulse.serial]
            qmpulse.swept_duration = dur
            # find all pulses that are connected to ``qmpulse`` and align them
            to_process = set(qmpulse.next_)
            while to_process:
                next_qmpulse = to_process.pop()
                to_process |= next_qmpulse.next_
                qmpulse.elements_to_align.add(next_qmpulse.element)
                next_qmpulse.wait_time -= qmpulse.wait_time + qmpulse.duration // 4

        _sweep_recursion(sweepers[1:], qubits, qmsequence, relaxation_time)



from dataclasses import dataclass

import numpy as np
from qibo import gates
from qibo.backends import NumpyBackend
from qibo.config import log, raise_error

from qibolab.native import NativeType
from qibolab.transpilers.abstract import Transpiler
from qibolab.transpilers.unitary_decompositions import (
    two_qubit_decomposition,
    u3_decomposition,
)

backend = NumpyBackend()


class GateDecompositions:
    """Abstract data structure that holds decompositions of gates."""

    def __init__(self):
        self.decompositions = {}

    def add(self, gate, decomposition):
        """Register a decomposition for a gate."""
        self.decompositions[gate] = decomposition

    def count_2q(self, gate):
        """Count the number of two-qubit gates in the decomposition of the given gate."""
        if gate.parameters:
            decomposition = self.decompositions[gate.__class__](gate)
        else:
            decomposition = self.decompositions[gate.__class__]
        return len(tuple(g for g in decomposition if len(g.qubits) > 1))

    def count_1q(self, gate):
        """Count the number of single qubit gates in the decomposition of the given gate."""
        if gate.parameters:
            decomposition = self.decompositions[gate.__class__](gate)
        else:
            decomposition = self.decompositions[gate.__class__]
        return len(tuple(g for g in decomposition if len(g.qubits) == 1))

    def __call__(self, gate):
        """Decompose a gate."""
        decomposition = self.decompositions[gate.__class__]
        if callable(decomposition):
            decomposition = decomposition(gate)
        return [g.on_qubits({i: q for i, q in enumerate(gate.qubits)}) for g in decomposition]


def translate_gate(gate, native_gates: NativeType):
    """Maps Qibo gates to a hardware native implementation.

    Args:
        gate (qibo.gates.abstract.Gate): gate to be decomposed.
        native_gates (list): List of two qubit native gates
            supported by the quantum hardware ("CZ" and/or "iSWAP").

    Returns:
        Shortest list of native gates
    """
    if isinstance(gate, (gates.M, gates.I, gates.Align)):
        return gate

    if len(gate.qubits) == 1:
        return onequbit_dec(gate)

    if native_gates is NativeType.CZ | NativeType.iSWAP:
        # Check for a special optimized decomposition.
        if gate.__class__ in opt_dec.decompositions:
            return opt_dec(gate)
        # Check if the gate has a CZ decomposition
        if not gate.__class__ in iswap_dec.decompositions:
            return cz_dec(gate)
        # Check the decomposition with less 2 qubit gates.
        else:
            if cz_dec.count_2q(gate) < iswap_dec.count_2q(gate):
                return cz_dec(gate)
            elif cz_dec.count_2q(gate) > iswap_dec.count_2q(gate):
                return iswap_dec(gate)
            # If equal check the decomposition with less 1 qubit gates.
            # This is never used for now but may be useful for future generalization
            elif cz_dec.count_1q(gate) < iswap_dec.count_1q(gate):  # pragma: no cover
                return cz_dec(gate)
            else:  # pragma: no cover
                return iswap_dec(gate)
    elif native_gates is NativeType.CZ:
        return cz_dec(gate)
    elif native_gates is NativeType.iSWAP:
        if gate.__class__ in iswap_dec.decompositions:
            return iswap_dec(gate)
        else:
            # First decompose into CZ
            cz_decomposed = cz_dec(gate)
            # Then CZ are decomposed into iSWAP
            iswap_decomposed = []
            for g in cz_decomposed:
                # Need recursive function as gates.Unitary is not in iswap_dec
                for g_translated in translate_gate(g, NativeType.iSWAP):
                    iswap_decomposed.append(g_translated)
            return iswap_decomposed
    else:  # pragma: no cover
        raise_error(NotImplementedError, "Use only CZ and/or iSWAP as native gates")


onequbit_dec = GateDecompositions()
onequbit_dec.add(gates.H, [gates.U3(0, 7 * np.pi / 2, np.pi, 0)])
onequbit_dec.add(gates.X, [gates.U3(0, np.pi, 0, np.pi)])
onequbit_dec.add(gates.Y, [gates.U3(0, np.pi, 0, 0)])
# apply virtually by changing ``phase`` instead of using pulses
onequbit_dec.add(gates.Z, [gates.Z(0)])
onequbit_dec.add(gates.S, [gates.RZ(0, np.pi / 2)])
onequbit_dec.add(gates.SDG, [gates.RZ(0, -np.pi / 2)])
onequbit_dec.add(gates.T, [gates.RZ(0, np.pi / 4)])
onequbit_dec.add(gates.TDG, [gates.RZ(0, -np.pi / 4)])
onequbit_dec.add(gates.RX, lambda gate: [gates.U3(0, gate.parameters[0], -np.pi / 2, np.pi / 2)])
onequbit_dec.add(gates.RY, lambda gate: [gates.U3(0, gate.parameters[0], 0, 0)])
# apply virtually by changing ``phase`` instead of using pulses
onequbit_dec.add(gates.RZ, lambda gate: [gates.RZ(0, gate.parameters[0])])
# apply virtually by changing ``phase`` instead of using pulses
onequbit_dec.add(gates.GPI2, lambda gate: [gates.GPI2(0, gate.parameters[0])])
# implemented as single RX90 pulse
onequbit_dec.add(gates.U1, lambda gate: [gates.RZ(0, gate.parameters[0])])
onequbit_dec.add(gates.U2, lambda gate: [gates.U3(0, np.pi / 2, gate.parameters[0], gate.parameters[1])])
onequbit_dec.add(gates.U3, lambda gate: [gates.U3(0, gate.parameters[0], gate.parameters[1], gate.parameters[2])])
onequbit_dec.add(
    gates.Unitary,
    lambda gate: [gates.U3(0, *u3_decomposition(gate.parameters[0]))],
)
onequbit_dec.add(
    gates.FusedGate,
    lambda gate: [gates.U3(0, *u3_decomposition(gate.matrix(backend)))],
)

# register the iSWAP decompositions
iswap_dec = GateDecompositions()
iswap_dec.add(
    gates.CNOT,
    [
        gates.U3(0, 7 * np.pi / 2, np.pi, 0),
        gates.U3(1, np.pi / 2, -np.pi, -np.pi),
        gates.iSWAP(0, 1),
        gates.U3(0, np.pi, 0, np.pi),
        gates.U3(1, np.pi / 2, -np.pi, -np.pi),
        gates.iSWAP(0, 1),
        gates.U3(0, np.pi / 2, np.pi / 2, -np.pi),
        gates.U3(1, np.pi / 2, -np.pi, -np.pi / 2),
    ],
)
iswap_dec.add(
    gates.CZ,
    [
        gates.U3(0, 7 * np.pi / 2, np.pi, 0),
        gates.U3(1, 7 * np.pi / 2, np.pi, 0),
        gates.U3(1, np.pi / 2, -np.pi, -np.pi),
        gates.iSWAP(0, 1),
        gates.U3(0, np.pi, 0, np.pi),
        gates.U3(1, np.pi / 2, -np.pi, -np.pi),
        gates.iSWAP(0, 1),
        gates.U3(0, np.pi / 2, np.pi / 2, -np.pi),
        gates.U3(1, np.pi / 2, -np.pi, -np.pi / 2),
        gates.U3(1, 7 * np.pi / 2, np.pi, 0),
    ],
)
iswap_dec.add(
    gates.SWAP,
    [
        gates.iSWAP(0, 1),
        gates.U3(1, np.pi / 2, -np.pi / 2, np.pi / 2),
        gates.iSWAP(0, 1),
        gates.U3(0, np.pi / 2, -np.pi / 2, np.pi / 2),
        gates.iSWAP(0, 1),
        gates.U3(1, np.pi / 2, -np.pi / 2, np.pi / 2),
    ],
)
iswap_dec.add(gates.iSWAP, [gates.iSWAP(0, 1)])

# register CZ decompositions
cz_dec = GateDecompositions()
cz_dec.add(gates.CNOT, [gates.H(1), gates.CZ(0, 1), gates.H(1)])
cz_dec.add(gates.CZ, [gates.CZ(0, 1)])
cz_dec.add(
    gates.SWAP,
    [
        gates.H(1),
        gates.CZ(0, 1),
        gates.H(1),
        gates.H(0),
        gates.CZ(1, 0),
        gates.H(0),
        gates.H(1),
        gates.CZ(0, 1),
        gates.H(1),
    ],
)
cz_dec.add(
    gates.iSWAP,
    [
        gates.U3(0, np.pi / 2.0, 0, -np.pi / 2.0),
        gates.U3(1, np.pi / 2.0, 0, -np.pi / 2.0),
        gates.CZ(0, 1),
        gates.H(0),
        gates.H(1),
        gates.CZ(0, 1),
        gates.H(0),
        gates.H(1),
    ],
)
cz_dec.add(
    gates.CRX,
    lambda gate: [
        gates.RX(1, gate.parameters[0] / 2.0),
        gates.CZ(0, 1),
        gates.RX(1, -gate.parameters[0] / 2.0),
        gates.CZ(0, 1),
    ],
)
cz_dec.add(
    gates.CRY,
    lambda gate: [
        gates.RY(1, gate.parameters[0] / 2.0),
        gates.CZ(0, 1),
        gates.RY(1, -gate.parameters[0] / 2.0),
        gates.CZ(0, 1),
    ],
)
cz_dec.add(
    gates.CRZ,
    lambda gate: [
        gates.RZ(1, gate.parameters[0] / 2.0),
        gates.H(1),
        gates.CZ(0, 1),
        gates.RX(1, -gate.parameters[0] / 2.0),
        gates.CZ(0, 1),
        gates.H(1),
    ],
)
cz_dec.add(
    gates.CU1,
    lambda gate: [
        gates.RZ(0, gate.parameters[0] / 2.0),
        gates.H(1),
        gates.CZ(0, 1),
        gates.RX(1, -gate.parameters[0] / 2.0),
        gates.CZ(0, 1),
        gates.H(1),
        gates.RZ(1, gate.parameters[0] / 2.0),
    ],
)
cz_dec.add(
    gates.CU2,
    lambda gate: [
        gates.RZ(1, (gate.parameters[1] - gate.parameters[0]) / 2.0),
        gates.H(1),
        gates.CZ(0, 1),
        gates.H(1),
        gates.U3(1, -np.pi / 4, 0, -(gate.parameters[1] + gate.parameters[0]) / 2.0),
        gates.H(1),
        gates.CZ(0, 1),
        gates.H(1),
        gates.U3(1, np.pi / 4, gate.parameters[0], 0),
    ],
)
cz_dec.add(
    gates.CU3,
    lambda gate: [
        gates.RZ(1, (gate.parameters[2] - gate.parameters[1]) / 2.0),
        gates.H(1),
        gates.CZ(0, 1),
        gates.H(1),
        gates.U3(1, -gate.parameters[0] / 2.0, 0, -(gate.parameters[2] + gate.parameters[1]) / 2.0),
        gates.H(1),
        gates.CZ(0, 1),
        gates.H(1),
        gates.U3(1, gate.parameters[0] / 2.0, gate.parameters[1], 0),
    ],
)
cz_dec.add(
    gates.FSWAP,
    [
        gates.U3(0, np.pi / 2, -np.pi / 2, -np.pi),
        gates.U3(1, np.pi / 2, np.pi / 2, np.pi / 2),
        gates.CZ(0, 1),
        gates.U3(0, np.pi / 2, 0, -np.pi / 2),
        gates.U3(1, np.pi / 2, 0, np.pi / 2),
        gates.CZ(0, 1),
        gates.U3(0, np.pi / 2, np.pi / 2, -np.pi),
        gates.U3(1, np.pi / 2, 0, -np.pi),
    ],
)
cz_dec.add(
    gates.RXX,
    lambda gate: [
        gates.H(0),
        gates.CZ(0, 1),
        gates.RX(1, gate.parameters[0]),
        gates.CZ(0, 1),
        gates.H(0),
    ],
)
cz_dec.add(
    gates.RYY,
    lambda gate: [
        gates.RX(0, np.pi / 2),
        gates.U3(1, np.pi / 2, np.pi / 2, -np.pi),
        gates.CZ(0, 1),
        gates.RX(1, gate.parameters[0]),
        gates.CZ(0, 1),
        gates.RX(0, -np.pi / 2),
        gates.U3(1, np.pi / 2, 0, np.pi / 2),
    ],
)
cz_dec.add(
    gates.RZZ,
    lambda gate: [
        gates.H(1),
        gates.CZ(0, 1),
        gates.RX(1, gate.parameters[0]),
        gates.CZ(0, 1),
        gates.H(1),
    ],
)
cz_dec.add(
    gates.TOFFOLI,
    [
        gates.CZ(1, 2),
        gates.RX(2, -np.pi / 4),
        gates.CZ(0, 2),
        gates.RX(2, np.pi / 4),
        gates.CZ(1, 2),
        gates.RX(2, -np.pi / 4),
        gates.CZ(0, 2),
        gates.RX(2, np.pi / 4),
        gates.RZ(1, np.pi / 4),
        gates.H(1),
        gates.CZ(0, 1),
        gates.RZ(0, np.pi / 4),
        gates.RX(1, -np.pi / 4),
        gates.CZ(0, 1),
        gates.H(1),
    ],
)
cz_dec.add(gates.Unitary, lambda gate: two_qubit_decomposition(0, 1, gate.parameters[0]))
cz_dec.add(gates.fSim, lambda gate: two_qubit_decomposition(0, 1, gate.matrix(backend)))
cz_dec.add(gates.GeneralizedfSim, lambda gate: two_qubit_decomposition(0, 1, gate.matrix(backend)))


# register other optimized gate decompositions
opt_dec = GateDecompositions()
opt_dec.add(
    gates.SWAP,
    [
        gates.H(0),
        gates.SDG(0),
        gates.SDG(1),
        gates.iSWAP(0, 1),
        gates.CZ(0, 1),
        gates.H(1),
    ],
)


@dataclass
class NativeGates(Transpiler):
    """Translates a circuit to native gates.

    Args:
        circuit (qibo.models.Circuit): Circuit model to translate into native gates.
        two_qubit_natives (list): List of two qubit native gates
            supported by the quantum hardware ("CZ" and/or "iSWAP").

    Returns:
        new (qibo.models.Circuit): Equivalent circuit with native gates.
    """

    two_qubit_natives: NativeType
    translate_single_qubit: bool = True
    verbose: bool = False

    def tlog(self, message):
        """Print messages only if ``verbose`` was set to ``True``."""
        # TODO: Move this in ``AbstractTranspiler``
        if self.verbose:
            log.info(message)

    def is_satisfied(self, circuit):
        """Checks if a circuit can be executed on tii5q.

        Args:
            circuit (qibo.models.Circuit): Circuit model to check.
            two_qubit_natives (list): List of two qubit native gates
                supported by the quantum hardware ("CZ" and/or "iSWAP").
            middle_qubit (int): Hardware middle qubit.
            verbose (bool): If ``True`` it prints debugging log messages.

        Returns ``True`` if the following conditions are satisfied:
            - Circuit does not contain more than two-qubit gates.
            - All one-qubit gates are I, Z, RZ or U3.
            - All two-qubit gates are CZ or iSWAP based on two_qubit_natives.
            - All two-qubit gates have qubit 0 as target or control.

            otherwise returns ``False``.
        """
        for gate in circuit.queue:
            if isinstance(gate, gates.M):
                continue

            if len(gate.qubits) == 1:
                # TODO: Make setting single-qubit native gates more flexible
                if not isinstance(gate, (gates.I, gates.Z, gates.RZ, gates.U3)):
                    self.tlog(f"{gate.name} is not a single qubit native gate.")
                    return False

            elif len(gate.qubits) == 2:
                if not (NativeType.from_gate(gate) in self.two_qubit_natives):
                    self.tlog(f"{gate.name} is not a two qubit native gate.")
                    return False

            else:
                self.tlog(f"{gate.name} acts on more than two qubits.")
                return False

        self.tlog("Circuit can be executed.")
        return True

    def __call__(self, circuit):
        new = circuit.__class__(circuit.nqubits)
        for gate in circuit.queue:
            if len(gate.qubits) > 1 or self.translate_single_qubit:
                new.add(translate_gate(gate, self.two_qubit_natives))
            else:
                new.add(gate)
        return new, list(range(circuit.nqubits))



"""Instrument for using the Zurich Instruments (Zhinst) devices."""

import copy
import os
from collections import defaultdict
from dataclasses import dataclass, replace
from typing import Tuple

import laboneq._token
import laboneq.simple as lo
import numpy as np
from laboneq.contrib.example_helpers.plotting.plot_helpers import plot_simulation
from laboneq.dsl.experiment.pulse_library import (
    sampled_pulse_complex,
    sampled_pulse_real,
)
from qibo.config import log

from qibolab import AcquisitionType, AveragingMode, ExecutionParameters
from qibolab.instruments.abstract import INSTRUMENTS_DATA_FOLDER, Controller
from qibolab.instruments.port import Port
from qibolab.pulses import FluxPulse, PulseSequence, PulseType
from qibolab.sweeper import Parameter

# this env var just needs to be set
os.environ["LABONEQ_TOKEN"] = "not required"
laboneq._token.is_valid_token = lambda _token: True  # pylint: disable=W0212

NANO_TO_SECONDS = 1e-9
SERVER_PORT = "8004"
COMPILER_SETTINGS = {
    "SHFSG_FORCE_COMMAND_TABLE": True,
    "SHFSG_MIN_PLAYWAVE_HINT": 32,
    "SHFSG_MIN_PLAYZERO_HINT": 32,
}

"""Translating to Zurich ExecutionParameters"""
ACQUISITION_TYPE = {
    AcquisitionType.INTEGRATION: lo.AcquisitionType.INTEGRATION,
    AcquisitionType.RAW: lo.AcquisitionType.RAW,
    AcquisitionType.DISCRIMINATION: lo.AcquisitionType.DISCRIMINATION,
}

AVERAGING_MODE = {
    AveragingMode.CYCLIC: lo.AveragingMode.CYCLIC,
    AveragingMode.SINGLESHOT: lo.AveragingMode.SINGLE_SHOT,
}

SWEEPER_SET = {"amplitude", "frequency", "duration", "relative_phase"}
SWEEPER_BIAS = {"bias"}
SWEEPER_START = {"start"}


def select_pulse(pulse, pulse_type):
    """Pulse translation"""

    if "IIR" not in str(pulse.shape):
        if str(pulse.shape) == "Rectangular()":
            can_compress = pulse.type is not PulseType.READOUT
            return lo.pulse_library.const(
                uid=(f"{pulse_type}_{pulse.qubit}_"),
                length=round(pulse.duration * NANO_TO_SECONDS, 9),
                amplitude=pulse.amplitude,
                can_compress=can_compress,
            )
        if "Gaussian" in str(pulse.shape):
            sigma = pulse.shape.rel_sigma
            return lo.pulse_library.gaussian(
                uid=(f"{pulse_type}_{pulse.qubit}_"),
                length=round(pulse.duration * NANO_TO_SECONDS, 9),
                amplitude=pulse.amplitude,
                sigma=2 / sigma,
                zero_boundaries=False,
            )

        if "GaussianSquare" in str(pulse.shape):
            sigma = pulse.shape.rel_sigma
            can_compress = pulse.type is not PulseType.READOUT
            return lo.pulse_library.gaussian_square(
                uid=(f"{pulse_type}_{pulse.qubit}_"),
                length=round(pulse.duration * NANO_TO_SECONDS, 9),
                width=round(pulse.duration * NANO_TO_SECONDS, 9) * 0.9,  # 90% Flat
                amplitude=pulse.amplitude,
                can_compress=can_compress,
                sigma=2 / sigma,
                zero_boundaries=False,
            )

        if "Drag" in str(pulse.shape):
            sigma = pulse.shape.rel_sigma
            beta = pulse.shape.beta
            return lo.pulse_library.drag(
                uid=(f"{pulse_type}_{pulse.qubit}_"),
                length=round(pulse.duration * NANO_TO_SECONDS, 9),
                amplitude=pulse.amplitude,
                sigma=2 / sigma,
                beta=beta,
                zero_boundaries=False,
            )

    if np.all(pulse.envelope_waveform_q.data == 0):
        return sampled_pulse_real(
            uid=(f"{pulse_type}_{pulse.qubit}_"),
            samples=pulse.envelope_waveform_i.data,
            can_compress=True,
        )
    else:
        # TODO: Test this when we have pulses that use it
        return sampled_pulse_complex(
            uid=(f"{pulse_type}_{pulse.qubit}_"),
            samples=pulse.envelope_waveform_i.data + (1j * pulse.envelope_waveform_q.data),
            can_compress=True,
        )

    # TODO: if "Slepian" in str(pulse.shape):
    # Implement Slepian shaped flux pulse https://arxiv.org/pdf/0909.5368.pdf

    # """
    # Typically, the sampler function should discard ``length`` and ``amplitude``, and
    # instead assume that the pulse extends from -1 to 1, and that it has unit
    # amplitude. LabOne Q will automatically rescale the sampler's output to the correct
    # amplitude and length.

    # They don't even do that on their notebooks
    # and just use lenght and amplitude but we have to check


@dataclass
class ZhPort(Port):
    name: Tuple[str, str]
    offset: float = 0.0
    power_range: int = 0


class ZhPulse:
    """Zurich pulse from qibolab pulse translation."""

    def __init__(self, pulse):
        """Zurich pulse from qibolab pulse"""
        self.pulse = pulse
        """Qibolab pulse"""
        self.signal = f"{pulse.type.name.lower()}{pulse.qubit}"
        """Line associated with the pulse"""
        self.zhpulse = select_pulse(pulse, pulse.type.name.lower())
        """Zurich pulse"""


class ZhSweeper:
    """
    Zurich sweeper from qibolab sweeper for pulse parameters
    Amplitude, Duration, Frequency (and maybe Phase)

    """

    def __init__(self, pulse, sweeper, qubit):
        self.sweeper = sweeper
        """Qibolab sweeper"""

        self.pulse = pulse
        """Qibolab pulse associated to the sweeper"""
        self.signal = f"{pulse.type.name.lower()}{pulse.qubit}"
        """Line associated with the pulse"""
        self.zhpulse = ZhPulse(pulse).zhpulse
        """Zurich pulse associated to the sweeper"""

        self.zhsweeper = self.select_sweeper(pulse.type, sweeper, qubit)
        """Zurich sweeper"""

        self.zhsweepers = [self.select_sweeper(pulse.type, sweeper, qubit)]
        """
        Zurich sweepers, Need something better to store multiple sweeps on the same pulse

        Not properly implemented as it was only used on Rabi amplitude vs lenght and it
        was an unused routine.
        """

    @staticmethod  # pylint: disable=R0903
    def select_sweeper(ptype, sweeper, qubit):
        """Sweeper translation"""

        if sweeper.parameter is Parameter.amplitude:
            return lo.SweepParameter(
                uid=sweeper.parameter.name,
                values=copy.copy(sweeper.values),
            )
        if sweeper.parameter is Parameter.duration:
            return lo.SweepParameter(
                uid=sweeper.parameter.name,
                values=sweeper.values * NANO_TO_SECONDS,
            )
        if sweeper.parameter is Parameter.relative_phase:
            return lo.SweepParameter(
                uid=sweeper.parameter.name,
                values=sweeper.values,
            )
        if sweeper.parameter is Parameter.frequency:
            if ptype is PulseType.READOUT:
                intermediate_frequency = qubit.readout_frequency - qubit.readout.local_oscillator.frequency
            elif ptype is PulseType.DRIVE:
                intermediate_frequency = qubit.drive_frequency - qubit.drive.local_oscillator.frequency
            return lo.LinearSweepParameter(
                uid=sweeper.parameter.name,
                start=sweeper.values[0] + intermediate_frequency,
                stop=sweeper.values[-1] + intermediate_frequency,
                count=len(sweeper.values),
            )

    def add_sweeper(self, sweeper, qubit):
        """Add sweeper to list of sweepers"""
        self.zhsweepers.append(self.select_sweeper(self.pulse.type, sweeper, qubit))


class ZhSweeperLine:
    """
    Zurich sweeper from qibolab sweeper for non pulse parameters
    Bias, Delay (, power_range, local_oscillator frequency, offset ???)

    For now Parameter.bias sweepers are implemented as Parameter.Amplitude
    on a flux pulse. We may want to keep this class separate for future
    Near Time sweeps
    """

    def __init__(self, sweeper, qubit=None, sequence=None):
        self.sweeper = sweeper
        """Qibolab sweeper"""

        if sweeper.parameter is Parameter.bias:
            pulse = FluxPulse(
                start=sequence.start,
                duration=sequence.duration,
                amplitude=1,
                shape="Rectangular",
                channel=qubit.flux.name,
                qubit=qubit.name,
            )
            self.pulse = pulse
            self.signal = f"flux{qubit.name}"

            self.zhpulse = lo.pulse_library.const(
                uid=(f"{pulse.type.name.lower()}_{pulse.qubit}_"),
                length=round(pulse.duration * NANO_TO_SECONDS, 9),
                amplitude=pulse.amplitude,
            )

        # Need something better to store multiple sweeps on the same pulse
        self.zhsweeper = self.select_sweeper(sweeper)

    @staticmethod  # pylint: disable=R0903
    def select_sweeper(sweeper):
        """Sweeper translation"""
        if sweeper.parameter is Parameter.bias:
            return lo.SweepParameter(
                uid=sweeper.parameter.name,
                values=sweeper.values,
            )
        if sweeper.parameter is Parameter.start:
            return lo.SweepParameter(
                uid=sweeper.parameter.name,
                values=sweeper.values * NANO_TO_SECONDS,
            )


class Zurich(Controller):
    """Zurich driver main class"""

    PortType = ZhPort

    def __init__(self, name, descriptor, use_emulation=False, time_of_flight=0.0, smearing=0.0):
        self.name = name
        "Setup name (str)"

        self.descriptor = descriptor
        """
        Port and device mapping in yaml text (str)

        It should be used as a template by adding extra lines for each of the different
        frequency pulses played thought the same port after parsing the sequence.
        """

        self.emulation = use_emulation
        "Enable emulation mode (bool)"
        self.is_connected = False
        "Is the device connected ? (bool)"

        self.signal_map = {}
        "Signals to lines mapping"
        self.calibration = lo.Calibration()
        "Zurich calibration object)"

        self.device_setup = None
        self.session = None
        self.device = None
        "Zurich device parameters for connection"

        self.time_of_flight = time_of_flight
        self.smearing = smearing
        self.chip = "iqm5q"
        "Parameters read from the runcard not part of ExecutionParameters"

        self.exp = None
        self.experiment = None
        self.exp_options = ExecutionParameters()
        self.exp_calib = lo.Calibration()
        self.results = None
        "Zurich experiment definitions"

        self.acquisition_type = None
        "To store if the AcquisitionType.SPECTROSCOPY needs to be enabled by parsing the sequence"

        self.sequence = defaultdict(list)
        "Zurich pulse sequence"
        self.sequence_qibo = None
        # Remove if able

        self.sweepers = []
        self.nt_sweeps = None
        "Storing sweepers"
        # Improve the storing of multiple sweeps
        self._ports = {}

    def connect(self):
        self.create_device_setup()
        # To fully remove logging #configure_logging=False
        # I strongly advise to set it to 20 to have time estimates of the experiment duration!
        self.session = lo.Session(self.device_setup, log_level=30)
        self.device = self.session.connect(do_emulation=self.emulation)
        self.is_connected = True

    def create_device_setup(self):
        """Loads the device setup to address the instruments"""
        self.device_setup = lo.DeviceSetup.from_dict(
            data=self.descriptor,
            server_host="localhost",
            server_port=SERVER_PORT,
            setup_name=self.name,
        )

    def start(self):
        """Empty method to comply with Instrument interface."""

    def stop(self):
        """Empty method to comply with Instrument interface."""

    def disconnect(self):
        if self.is_connected:
            self.device = self.session.disconnect()
            self.is_connected = False

    def setup(self, *args, **kwargs):
        """Empty method to comply with Instrument interface."""

    def calibration_step(self, qubits, options):
        """
        Zurich general pre experiment calibration definitions

        Change to get frequencies from sequence
        """

        for qubit in qubits.values():
            if qubit.flux_coupler:
                self.register_flux_line(qubit)
            else:
                if qubit.flux is not None:
                    self.register_flux_line(qubit)
                if len(self.sequence[f"drive{qubit.name}"]) != 0:
                    self.register_drive_line(
                        qubit=qubit,
                        intermediate_frequency=qubit.drive_frequency - qubit.drive.local_oscillator.frequency,
                    )
                if len(self.sequence[f"readout{qubit.name}"]) != 0:
                    self.register_readout_line(
                        qubit=qubit,
                        intermediate_frequency=qubit.readout_frequency - qubit.readout.local_oscillator.frequency,
                    )
                    if options.fast_reset is not False:
                        if len(self.sequence[f"drive{qubit.name}"]) == 0:
                            self.register_drive_line(
                                qubit=qubit,
                                intermediate_frequency=qubit.drive_frequency - qubit.drive.local_oscillator.frequency,
                            )
        self.device_setup.set_calibration(self.calibration)

    def register_readout_line(self, qubit, intermediate_frequency):
        """Registers qubit measure and acquire lines to calibration and signal map.

        Note
        ----
        To allow debugging with and oscilloscope, just set the following::

            self.calibration[f"/logical_signal_groups/q{q}/measure_line"] = lo.SignalCalibration(
                ...,
                local_oscillator=lo.Oscillator(
                    ...
                    frequency=0.0,
                ),
                ...,
                port_mode=lo.PortMode.LF,
                ...,
            )

        """

        q = qubit.name  # pylint: disable=C0103
        self.signal_map[f"measure{q}"] = self.device_setup.logical_signal_groups[f"q{q}"].logical_signals[
            "measure_line"
        ]
        self.calibration[f"/logical_signal_groups/q{q}/measure_line"] = lo.SignalCalibration(
            oscillator=lo.Oscillator(
                frequency=intermediate_frequency,
                modulation_type=lo.ModulationType.SOFTWARE,
            ),
            local_oscillator=lo.Oscillator(
                uid="lo_shfqa",
                frequency=int(qubit.readout.local_oscillator.frequency),
            ),
            range=qubit.readout.power_range,
            port_delay=None,
            delay_signal=0,
        )

        self.signal_map[f"acquire{q}"] = self.device_setup.logical_signal_groups[f"q{q}"].logical_signals[
            "acquire_line"
        ]
        self.calibration[f"/logical_signal_groups/q{q}/acquire_line"] = lo.SignalCalibration(
            oscillator=lo.Oscillator(
                frequency=intermediate_frequency,
                modulation_type=lo.ModulationType.SOFTWARE,
            ),
            local_oscillator=lo.Oscillator(
                uid="lo_shfqa",
                frequency=int(qubit.readout.local_oscillator.frequency),
            ),
            range=qubit.feedback.power_range,
            port_delay=self.time_of_flight * NANO_TO_SECONDS,
            threshold=qubit.threshold,
        )

    def register_drive_line(self, qubit, intermediate_frequency):
        """Registers qubit drive line to calibration and signal map."""
        q = qubit.name  # pylint: disable=C0103
        self.signal_map[f"drive{q}"] = self.device_setup.logical_signal_groups[f"q{q}"].logical_signals["drive_line"]
        self.calibration[f"/logical_signal_groups/q{q}/drive_line"] = lo.SignalCalibration(
            oscillator=lo.Oscillator(
                frequency=intermediate_frequency,
                modulation_type=lo.ModulationType.HARDWARE,
            ),
            local_oscillator=lo.Oscillator(
                uid="lo_shfqc",
                frequency=int(qubit.drive.local_oscillator.frequency),
            ),
            range=qubit.drive.power_range,
            port_delay=None,
            delay_signal=0,
        )

    def register_flux_line(self, qubit):
        """Registers qubit flux line to calibration and signal map."""
        q = qubit.name  # pylint: disable=C0103
        self.signal_map[f"flux{q}"] = self.device_setup.logical_signal_groups[f"q{q}"].logical_signals["flux_line"]
        self.calibration[f"/logical_signal_groups/q{q}/flux_line"] = lo.SignalCalibration(
            range=qubit.flux.power_range,
            port_delay=None,
            delay_signal=0,
            voltage_offset=qubit.flux.offset,
        )

    def run_exp(self):
        """Compilation settings, compilation step, execution step and data retrival"""
        self.exp = self.session.compile(self.experiment, compiler_settings=COMPILER_SETTINGS)
        self.results = self.session.run(self.exp)

    @staticmethod
    def frequency_from_pulses(qubits, sequence):
        """Gets the frequencies from the pulses to the qubits"""
        # FIXME: Dual drive frequency experiments
        for pulse in sequence:
            qubit = qubits[pulse.qubit]
            if pulse.type is PulseType.READOUT:
                qubit.readout_frequency = pulse.frequency
            if pulse.type is PulseType.DRIVE:
                qubit.drive_frequency = pulse.frequency

    def experiment_flow(self, qubits, sequence, options, sweepers=[]):
        """
        Create the experiment object for the devices, following the steps separated one on each method:
        Translation, Calibration, Experiment Definition.
        """
        self.sequence_zh(sequence, qubits, sweepers)
        self.calibration_step(qubits, options)
        self.create_exp(qubits, options)

    # pylint: disable=W0221
    def play(self, qubits, sequence, options):
        """Play pulse sequence"""
        self.signal_map = {}
        dimensions = []
        if options.averaging_mode is AveragingMode.SINGLESHOT:
            dimensions = [options.nshots]

        self.frequency_from_pulses(qubits, sequence)

        self.experiment_flow(qubits, sequence, options)
        self.run_exp()

        # Get the results back
        results = {}
        for qubit in qubits.values():
            if qubit.flux_coupler:
                continue
            q = qubit.name  # pylint: disable=C0103
            if len(self.sequence[f"readout{q}"]) != 0:
                for i, ropulse in enumerate(self.sequence[f"readout{q}"]):
                    exp_res = self.results.get_data(f"sequence{q}_{i}")
                    if options.acquisition_type is AcquisitionType.DISCRIMINATION:
                        data = (
                            np.array([exp_res]) if options.averaging_mode is AveragingMode.CYCLIC else np.array(exp_res)
                        )
                        results[ropulse.pulse.serial] = options.results_type(data)
                        results[ropulse.pulse.qubit] = options.results_type(data)
                    else:
                        results[ropulse.pulse.serial] = options.results_type(data=np.array(exp_res))
                        results[ropulse.pulse.qubit] = options.results_type(data=np.array(exp_res))

        exp_dimensions = list(np.array(exp_res).shape)
        if dimensions != exp_dimensions:
            log.warning("dimensions %d , exp_dimensions %d", dimensions, exp_dimensions)
            log.warning("dimensions not properly ordered")

        # html containing the pulse sequence schedule
        # lo.show_pulse_sheet("pulses", self.exp)
        return results

    def sequence_zh(self, sequence, qubits, sweepers):
        """Qibo sequence to Zurich sequence"""

        # Define and assign the sequence
        zhsequence = defaultdict(list)
        self.sequence_qibo = sequence

        # Fill the sequences with pulses according to their lines in temporal order
        for pulse in sequence:
            zhsequence[f"{pulse.type.name.lower()}{pulse.qubit}"].append(ZhPulse(pulse))

        # Mess that gets the sweeper and substitutes the pulse it sweeps in the right place

        def nt_loop(sweeper):
            if len(sweepers) == 2:
                if not self.nt_sweeps:
                    self.nt_sweeps = [sweeper]
                else:
                    self.nt_sweeps.append(sweeper)
                self.sweepers.remove(sweeper)

        for sweeper in sweepers:
            if sweeper.parameter.name in SWEEPER_SET:
                for pulse in sweeper.pulses:
                    aux_list = zhsequence[f"{pulse.type.name.lower()}{pulse.qubit}"]
                    if sweeper.parameter is Parameter.frequency and pulse.type is PulseType.READOUT:
                        self.acquisition_type = lo.AcquisitionType.SPECTROSCOPY
                    if sweeper.parameter is Parameter.amplitude and pulse.type is PulseType.READOUT:
                        self.acquisition_type = lo.AcquisitionType.SPECTROSCOPY
                        nt_loop(sweeper)
                    for element in aux_list:
                        if pulse == element.pulse:
                            if isinstance(aux_list[aux_list.index(element)], ZhPulse):
                                aux_list[aux_list.index(element)] = ZhSweeper(pulse, sweeper, qubits[pulse.qubit])
                            elif isinstance(aux_list[aux_list.index(element)], ZhSweeper):
                                aux_list[aux_list.index(element)].add_sweeper(sweeper, qubits[pulse.qubit])

            if sweeper.parameter.name in SWEEPER_BIAS:
                for qubit in sweeper.qubits:
                    zhsequence[f"flux{qubit.name}"] = [ZhSweeperLine(sweeper, qubit, sequence)]

            # FIXME: This may not place the Zhsweeper when the start occurs among different sections or lines
            if sweeper.parameter.name in SWEEPER_START:
                pulse = sweeper.pulses[0]
                aux_list = zhsequence[f"{pulse.type.name.lower()}{pulse.qubit}"]
                for element in aux_list:
                    if pulse == element.pulse:
                        if isinstance(aux_list[aux_list.index(element)], ZhPulse):
                            aux_list.insert(
                                aux_list.index(element),
                                ZhSweeperLine(sweeper, pulse.qubit, sequence),
                            )
                            break

        self.sequence = zhsequence

    def create_exp(self, qubits, options):
        """Zurich experiment initialization usig their Experiment class"""

        # Setting experiment signal lines
        signals = []
        for qubit in qubits.values():
            q = qubit.name  # pylint: disable=C0103
            if qubit.flux_coupler:
                signals.append(lo.ExperimentSignal(f"flux{q}"))
            else:
                if len(self.sequence[f"drive{q}"]) != 0:
                    signals.append(lo.ExperimentSignal(f"drive{q}"))
                if qubit.flux is not None:
                    signals.append(lo.ExperimentSignal(f"flux{q}"))
                if len(self.sequence[f"readout{q}"]) != 0:
                    signals.append(lo.ExperimentSignal(f"measure{q}"))
                    signals.append(lo.ExperimentSignal(f"acquire{q}"))
                    if options.fast_reset is not False:
                        if len(self.sequence[f"drive{q}"]) == 0:
                            signals.append(lo.ExperimentSignal(f"drive{q}"))

        exp = lo.Experiment(
            uid="Sequence",
            signals=signals,
        )

        if self.acquisition_type:
            acquisition_type = self.acquisition_type
            self.acquisition_type = None
        else:
            acquisition_type = ACQUISITION_TYPE[options.acquisition_type]
        averaging_mode = AVERAGING_MODE[options.averaging_mode]
        exp_options = replace(options, acquisition_type=acquisition_type, averaging_mode=averaging_mode)

        exp_calib = lo.Calibration()
        # Near Time recursion loop or directly to Real Time recursion loop
        if self.nt_sweeps is not None:
            self.sweep_recursion_nt(qubits, exp_options, exp, exp_calib)
        else:
            self.define_exp(qubits, exp_options, exp, exp_calib)

    def define_exp(self, qubits, exp_options, exp, exp_calib):
        """Real time definition"""
        with exp.acquire_loop_rt(
            uid="shots",
            count=exp_options.nshots,
            acquisition_type=exp_options.acquisition_type,
            averaging_mode=exp_options.averaging_mode,
        ):
            # Recursion loop for sweepers or just play a sequence
            if len(self.sweepers) > 0:
                self.sweep_recursion(qubits, exp, exp_calib, exp_options)
            else:
                self.select_exp(exp, qubits, exp_options)
            exp.set_calibration(exp_calib)
            exp.set_signal_map(self.signal_map)
            self.experiment = exp

    def select_exp(self, exp, qubits, exp_options):
        """Build Zurich Experiment selecting the relevant sections"""
        if "drive" in str(self.sequence):
            if "flux" in str(self.sequence):
                self.flux(exp, qubits)
                self.drive(exp, qubits)
            else:
                self.drive(exp, qubits)
        elif "flux" in str(self.sequence):
            self.flux(exp, qubits)
        self.measure_relax(exp, qubits, exp_options.relaxation_time, exp_options.acquisition_type)
        if exp_options.fast_reset is not False:
            self.fast_reset(exp, qubits, exp_options.fast_reset)

    @staticmethod
    def play_sweep_select_single(exp, qubit, pulse, section, parameters, partial_sweep):
        """Play Zurich pulse when a single sweeper is involved"""
        if any("amplitude" in param for param in parameters):
            pulse.zhpulse.amplitude *= max(pulse.zhsweeper.values)
            pulse.zhsweeper.values /= max(pulse.zhsweeper.values)
            exp.play(
                signal=f"{section}{qubit.name}",
                pulse=pulse.zhpulse,
                amplitude=pulse.zhsweeper,
                phase=pulse.pulse.relative_phase,
            )
        elif any("duration" in param for param in parameters):
            exp.play(
                signal=f"{section}{qubit.name}",
                pulse=pulse.zhpulse,
                length=pulse.zhsweeper,
                phase=pulse.pulse.relative_phase,
            )
        elif any("relative_phase" in param for param in parameters):
            exp.play(
                signal=f"{section}{qubit.name}",
                pulse=pulse.zhpulse,
                phase=pulse.zhsweeper,  # I believe this is the global phase sweep
                # increment_oscillator_phase=pulse.zhsweeper, # I believe this is the relative phase sweep
            )
        elif "frequency" in partial_sweep.uid or partial_sweep.uid == "start":
            exp.play(
                signal=f"{section}{qubit.name}",
                pulse=pulse.zhpulse,
                phase=pulse.pulse.relative_phase,
            )

    # FIXME: Now hardcoded for the flux pulse for 2q gates
    @staticmethod
    def play_sweep_select_dual(exp, qubit, pulse, section, parameters):
        """Play Zurich pulse when a two sweepers are involved on the same pulse"""
        if "amplitude" in parameters and "duration" in parameters:
            for sweeper in pulse.zhsweepers:
                if sweeper.uid == "amplitude":
                    sweeper_amp_index = pulse.zhsweepers.index(sweeper)
                    sweeper.values = sweeper.values.copy()
                    pulse.zhpulse.amplitude *= max(abs(sweeper.values))
                    sweeper.values /= max(abs(sweeper.values))
                else:
                    sweeper_dur_index = pulse.zhsweepers.index(sweeper)

            exp.play(
                signal=f"{section}{qubit.name}",
                pulse=pulse.zhpulse,
                amplitude=pulse.zhsweepers[sweeper_amp_index],
                length=pulse.zhsweepers[sweeper_dur_index],
            )

    def play_sweep(self, exp, qubit, pulse, section):
        """Takes care of playing the sweepers and involved pulses for different options"""

        if isinstance(pulse, ZhSweeperLine):
            if pulse.zhsweeper.uid == "bias":
                exp.play(
                    signal=f"{section}{qubit.name}",
                    pulse=pulse.zhpulse,
                    amplitude=pulse.zhsweeper,
                )
        else:
            parameters = []
            for partial_sweep in pulse.zhsweepers:
                parameters.append(partial_sweep.uid)
            # Recheck partial sweeps
            if len(parameters) == 2:
                self.play_sweep_select_dual(exp, qubit, pulse, section, parameters)
            else:
                self.play_sweep_select_single(exp, qubit, pulse, section, parameters, partial_sweep)

    def flux(self, exp, qubits):
        """qubit flux or qubit coupler flux for bias sweep or pulses"""
        for qubit in qubits.values():
            q = qubit.name  # pylint: disable=C0103
            with exp.section(uid=f"sequence_bias{q}"):
                i = 0
                time = 0
                for pulse in self.sequence[f"flux{q}"]:
                    if not isinstance(pulse, ZhSweeperLine):
                        pulse.zhpulse.uid += str(i)
                        exp.delay(
                            signal=f"flux{q}",
                            time=round(pulse.pulse.start * NANO_TO_SECONDS, 9) - time,
                        )
                        time = round(pulse.pulse.duration * NANO_TO_SECONDS, 9) + round(
                            pulse.pulse.start * NANO_TO_SECONDS, 9
                        )
                    if isinstance(pulse, ZhSweeperLine):
                        self.play_sweep(exp, qubit, pulse, section="flux")
                    elif isinstance(pulse, ZhSweeper):
                        self.play_sweep(exp, qubit, pulse, section="flux")
                    elif isinstance(pulse, ZhPulse):
                        exp.play(signal=f"flux{q}", pulse=pulse.zhpulse)
                    i += 1

    def drive(self, exp, qubits):
        """qubit driving pulses"""
        for qubit in qubits.values():
            if qubit.flux_coupler:
                continue
            q = qubit.name  # pylint: disable=C0103
            time = 0
            i = 0
            if len(self.sequence[f"drive{q}"]) != 0:
                with exp.section(uid=f"sequence_drive{q}"):
                    for pulse in self.sequence[f"drive{q}"]:
                        if not isinstance(pulse, ZhSweeperLine):
                            exp.delay(
                                signal=f"drive{q}",
                                time=round(pulse.pulse.start * NANO_TO_SECONDS, 9) - time,
                            )
                            time = round(pulse.pulse.duration * NANO_TO_SECONDS, 9) + round(
                                pulse.pulse.start * NANO_TO_SECONDS, 9
                            )
                            pulse.zhpulse.uid += str(i)
                            if isinstance(pulse, ZhSweeper):
                                self.play_sweep(exp, qubit, pulse, section="drive")
                            elif isinstance(pulse, ZhPulse):
                                exp.play(
                                    signal=f"drive{q}",
                                    pulse=pulse.zhpulse,
                                    phase=pulse.pulse.relative_phase,
                                )
                                i += 1
                        elif isinstance(pulse, ZhSweeperLine):
                            exp.delay(signal=f"drive{q}", time=pulse.zhsweeper)

                    # TODO: Patch for T1 start, general ?
                    if len(self.sequence[f"readout{q}"]) > 0 and isinstance(
                        self.sequence[f"readout{q}"][0], ZhSweeperLine
                    ):
                        exp.delay(signal=f"drive{q}", time=self.sequence[f"readout{q}"][0].zhsweeper)
                        self.sequence[f"readout{q}"].remove(self.sequence[f"readout{q}"][0])

    @staticmethod
    def play_after_set(sequence, ptype):
        """Selects after which section the measurement goes"""
        longest = 0
        for pulse in sequence:
            if longest < pulse.finish:
                longest = pulse.finish
                qubit_after = pulse.qubit
        return f"sequence_{ptype}{qubit_after}"

    # For pulsed spectroscopy, set integration_length and either measure_pulse or measure_pulse_length.
    # For CW spectroscopy, set only integration_length and do not specify the measure signal.
    # For all other measurements, set either length or pulse for both the measure pulse and integration kernel.
    def measure_relax(self, exp, qubits, relaxation_time, acquisition_type):
        """qubit readout pulse, data acquisition and qubit relaxation"""
        play_after = None
        if len(self.sequence_qibo.qf_pulses) != 0 and len(self.sequence_qibo.qd_pulses) != 0:
            play_after = (
                self.play_after_set(self.sequence_qibo.qf_pulses, "bias")
                if self.sequence_qibo.qf_pulses.finish > self.sequence_qibo.qd_pulses.finish
                else self.play_after_set(self.sequence_qibo.qd_pulses, "drive")
            )
        elif len(self.sequence_qibo.qf_pulses) != 0:
            play_after = self.play_after_set(self.sequence_qibo.qf_pulses, "bias")
        elif len(self.sequence_qibo.qd_pulses) != 0:
            play_after = self.play_after_set(self.sequence_qibo.qd_pulses, "drive")

        readout_schedule = defaultdict(list)
        qubit_readout_schedule = defaultdict(list)
        iq_angle_readout_schedule = defaultdict(list)
        for qubit in qubits.values():
            if qubit.flux_coupler:
                continue
            q = qubit.name  # pylint: disable=C0103
            iq_angle = qubit.iq_angle
            if len(self.sequence[f"readout{q}"]) != 0:
                for i, pulse in enumerate(self.sequence[f"readout{q}"]):
                    readout_schedule[i].append(pulse)
                    qubit_readout_schedule[i].append(q)
                    iq_angle_readout_schedule[i].append(iq_angle)

        for i, (pulses, qubits, iq_angles) in enumerate(
            zip(readout_schedule.values(), qubit_readout_schedule.values(), iq_angle_readout_schedule.values())
        ):
            if i != 0:
                play_after = f"sequence_measure_{i-1}"
            # Section on the outside loop allows for multiplex
            with exp.section(uid=f"sequence_measure_{i}", play_after=play_after):
                for pulse, q, iq_angle in zip(pulses, qubits, iq_angles):
                    pulse.zhpulse.uid += str(i)

                    # Integration weights definition or load from the chip folder
                    weights_file = (
                        INSTRUMENTS_DATA_FOLDER / f"{self.chip}/weights/integration_weights_optimization_qubit_{q}.npy"
                    )
                    if weights_file.is_file():
                        log.info("I'm using optimized IW")
                        samples = np.load(
                            weights_file,
                            allow_pickle=True,
                        )
                        if acquisition_type == lo.AcquisitionType.DISCRIMINATION:
                            weight = lo.pulse_library.sampled_pulse_complex(
                                uid="weight" + pulse.zhpulse.uid,
                                # samples=samples[0] * np.exp(1j * qubit.iq_angle),
                                samples=samples[0] * np.exp(1j * iq_angle),
                            )
                        else:
                            weight = lo.pulse_library.sampled_pulse_complex(
                                uid="weight" + pulse.zhpulse.uid,
                                samples=samples[0],
                            )
                    else:
                        log.info("I'm using dumb IW")
                        # We adjust for smearing and remove smearing/2 at the end
                        exp.delay(
                            signal=f"acquire{q}",
                            time=self.smearing * NANO_TO_SECONDS,
                        )
                        if acquisition_type == lo.AcquisitionType.DISCRIMINATION:
                            weight = lo.pulse_library.sampled_pulse_complex(
                                np.ones([int(pulse.pulse.duration * 2 - 3 * self.smearing * NANO_TO_SECONDS)])
                                * np.exp(1j * iq_angle)
                            )
                        else:
                            weight = lo.pulse_library.const(
                                uid="weight" + pulse.zhpulse.uid,
                                length=round(pulse.pulse.duration * NANO_TO_SECONDS, 9)
                                - 1.5 * self.smearing * NANO_TO_SECONDS,
                                amplitude=1,
                            )

                    measure_pulse_parameters = {"phase": 0}

                    if i == len(self.sequence[f"readout{q}"]) - 1:
                        reset_delay = relaxation_time * NANO_TO_SECONDS
                    else:
                        # FIXME: Here time of flight or not ?
                        reset_delay = 0  # self.time_of_flight * NANO_TO_SECONDS

                    exp.measure(
                        acquire_signal=f"acquire{q}",
                        handle=f"sequence{q}_{i}",
                        integration_kernel=weight,
                        integration_kernel_parameters=None,
                        integration_length=None,
                        measure_signal=f"measure{q}",
                        measure_pulse=pulse.zhpulse,
                        measure_pulse_length=round(pulse.pulse.duration * NANO_TO_SECONDS, 9),
                        measure_pulse_parameters=measure_pulse_parameters,
                        measure_pulse_amplitude=None,
                        acquire_delay=self.time_of_flight * NANO_TO_SECONDS,
                        # reset_delay=relaxation_time * NANO_TO_SECONDS,
                        reset_delay=reset_delay,
                    )

    def fast_reset(self, exp, qubits, fast_reset):
        """
        Conditional fast reset after readout - small delay for signal processing
        This is a very naive approach that can be improved by repeating this step until
        we reach non fast reset fidelity
        https://quantum-computing.ibm.com/lab/docs/iql/manage/systems/reset/backend_reset
        """
        log.warning("Im fast resetting")
        for qubit_name in self.sequence_qibo.qubits:
            qubit = qubits[qubit_name]
            if qubit.flux_coupler:
                continue
            q = qubit.name  # pylint: disable=C0103
            with exp.section(uid=f"fast_reset{q}", play_after=f"sequence_measure"):
                with exp.match_local(handle=f"sequence{q}"):
                    with exp.case(state=0):
                        pass
                    with exp.case(state=1):
                        pulse = ZhPulse(qubit.native_gates.RX.pulse(0, 0))
                        exp.play(signal=f"drive{q}", pulse=pulse.zhpulse)

    @staticmethod
    def rearrange_sweepers(sweepers):
        """Rearranges sweepers from qibocal based on device hardware limitations"""
        rearranging_axes = [[], []]
        if len(sweepers) == 2:
            if sweepers[1].parameter is Parameter.frequency:
                if sweepers[0].parameter is Parameter.bias:
                    rearranging_axes[0] += [sweepers.index(sweepers[1])]
                    rearranging_axes[1] += [0]
                    sweeper_changed = sweepers[1]
                    sweepers.remove(sweeper_changed)
                    sweepers.insert(0, sweeper_changed)
                    log.warning("Sweepers were reordered")
                elif (
                    not sweepers[0].parameter is Parameter.amplitude
                    and sweepers[0].pulses[0].type is not PulseType.READOUT
                ):
                    rearranging_axes[0] += [sweepers.index(sweepers[1])]
                    rearranging_axes[1] += [0]
                    sweeper_changed = sweepers[1]
                    sweepers.remove(sweeper_changed)
                    sweepers.insert(0, sweeper_changed)
                    log.warning("Sweepers were reordered")
        return rearranging_axes, sweepers

    def offsets_off(self):
        """Sets the offsets from the HDAWGs to 0 after each experiment"""
        for sigout in range(0, 8):
            self.session.devices["device_hdawg"].awgs[0].sigouts[sigout].offset = 0
        self.session.devices["device_hdawg2"].awgs[0].sigouts[0].offset = 0

    def sweep(self, qubits, sequence: PulseSequence, options, *sweepers):
        """Play pulse and sweepers sequence"""

        self.signal_map = {}

        self.nt_sweeps = None

        sweepers = list(sweepers)

        dimensions = []
        if options.averaging_mode is AveragingMode.SINGLESHOT:
            dimensions = [options.nshots]

        for sweeper in sweepers:
            dimensions.append(len(sweeper.values))

        rearranging_axes, sweepers = self.rearrange_sweepers(sweepers)
        self.sweepers = sweepers

        self.frequency_from_pulses(qubits, sequence)

        self.experiment_flow(qubits, sequence, options, sweepers)
        self.run_exp()

        #  Get the results back
        results = {}
        for qubit in qubits.values():
            if qubit.flux_coupler:
                continue
            q = qubit.name  # pylint: disable=C0103
            if len(self.sequence[f"readout{q}"]) != 0:
                for i in range(len(self.sequence[f"readout{q}"])):
                    exp_res = self.results.get_data(f"sequence{q}_{i}")
                    # Reorder dimensions
                    exp_res = np.moveaxis(exp_res, rearranging_axes[0], rearranging_axes[1])
                    if options.acquisition_type is AcquisitionType.DISCRIMINATION:
                        data = (
                            np.array([exp_res]) if options.averaging_mode is AveragingMode.CYCLIC else np.array(exp_res)
                        )
                        data = data.real
                        results[self.sequence[f"readout{q}"][i].pulse.serial] = options.results_type(data)
                        results[self.sequence[f"readout{q}"][i].pulse.qubit] = options.results_type(data)
                    else:
                        results[self.sequence[f"readout{q}"][i].pulse.serial] = options.results_type(
                            data=np.array(exp_res)
                        )
                        results[self.sequence[f"readout{q}"][i].pulse.qubit] = options.results_type(
                            data=np.array(exp_res)
                        )

        exp_dimensions = list(np.array(exp_res).shape)
        if dimensions != exp_dimensions:
            log.warning("dimensions {}, exp_dimensions {}".format(dimensions, exp_dimensions))
            log.warning("dimensions not properly ordered")

        self.offsets_off()

        # html containing the pulse sequence schedule
        # lo.show_pulse_sheet("pulses", self.exp)
        return results

    def sweep_recursion(self, qubits, exp, exp_calib, exp_options):
        """Sweepers recursion for multiple nested Real Time sweepers"""

        sweeper = self.sweepers[0]

        i = len(self.sweepers) - 1
        self.sweepers.remove(sweeper)
        parameter = None

        if sweeper.parameter is Parameter.frequency:
            for pulse in sweeper.pulses:
                line = "drive" if pulse.type is PulseType.DRIVE else "measure"
                zhsweeper = ZhSweeper(pulse, sweeper, qubits[sweeper.pulses[0].qubit]).zhsweeper
                zhsweeper.uid = "frequency"  # TODO: Changing the name from "frequency" breaks it f"frequency_{i}
                exp_calib[f"{line}{pulse.qubit}"] = lo.SignalCalibration(
                    oscillator=lo.Oscillator(
                        frequency=zhsweeper,
                        modulation_type=lo.ModulationType.HARDWARE,
                    )
                )
        if sweeper.parameter is Parameter.amplitude:
            for pulse in sweeper.pulses:
                pulse = pulse.copy()
                pulse.amplitude *= max(abs(sweeper.values))

                # FIXME: Proper copy(sweeper) here
                # sweeper_aux = copy.copy(sweeper)
                aux_max = max(abs(sweeper.values))

                sweeper.values /= aux_max
                parameter = ZhSweeper(pulse, sweeper, qubits[sweeper.pulses[0].qubit]).zhsweeper
                sweeper.values *= aux_max

        if sweeper.parameter is Parameter.bias:
            for qubit in sweeper.qubits:
                parameter = ZhSweeperLine(sweeper, qubit, self.sequence_qibo).zhsweeper

        elif sweeper.parameter is Parameter.start:
            parameter = ZhSweeperLine(sweeper).zhsweeper

        elif parameter is None:
            parameter = ZhSweeper(sweeper.pulses[0], sweeper, qubits[sweeper.pulses[0].qubit]).zhsweeper

        with exp.sweep(
            uid=f"sweep_{sweeper.parameter.name.lower()}_{i}",  # FIXME: This uid trouble double freq ???
            parameter=parameter,
            reset_oscillator_phase=True,  # FIXME: Should we reset this phase ???
        ):
            if len(self.sweepers) > 0:
                self.sweep_recursion(qubits, exp, exp_calib, exp_options)
            else:
                self.select_exp(exp, qubits, exp_options)

    def sweep_recursion_nt(self, qubits, options, exp, exp_calib):
        """
        Sweepers recursion for Near Time sweepers. Faster than regular software sweepers as
        they are executed on the actual device by (software ? or slower hardware ones)

        You want to avoid them so for now they are implement for a specific sweep.
        """

        log.info("nt Loop")

        sweeper = self.nt_sweeps[0]

        i = len(self.nt_sweeps) - 1
        self.nt_sweeps.remove(sweeper)

        parameter = None

        if sweeper.parameter is Parameter.amplitude:
            for pulse in sweeper.pulses:
                pulse = pulse.copy()
                pulse.amplitude *= max(abs(sweeper.values))

                # FIXME: Proper copy(sweeper) here
                # sweeper_aux = copy.copy(sweeper)
                aux_max = max(abs(sweeper.values))

                sweeper.values /= aux_max
                zhsweeper = ZhSweeper(pulse, sweeper, qubits[sweeper.pulses[0].qubit]).zhsweeper
                sweeper.values *= aux_max

                zhsweeper.uid = "amplitude"  # f"amplitude{i}"
                path = "DEV12146"  # Hardcoded for SHFQC(SHFQA)
                parameter = zhsweeper

        elif parameter is None:
            parameter = ZhSweeper(sweeper.pulses[0], sweeper, qubits[sweeper.pulses[0].qubit]).zhsweeper

        with exp.sweep(
            uid=f"sweep_{sweeper.parameter.name.lower()}_{i}",
            parameter=parameter,
        ):
            exp.set_node(
                path=f"/{path}/qachannels/*/oscs/0/gain",  # Hardcoded SHFQA device
                value=parameter,
            )

            if len(self.nt_sweeps) > 0:
                self.sweep_recursion_nt(qubits, options, exp, exp_calib)
            else:
                self.define_exp(qubits, options, exp, exp_calib)

    def play_sequences(self, qubits, sequence, options):
        pass

    # -----------------------------------------------------------------------------

    def play_sim(self, qubits, sequence, options, sim_time):
        """Play pulse sequence"""

        self.experiment_flow(qubits, sequence, options)
        self.run_sim(sim_time)

    # TODO: Implement further pulse viewing functions from 2.2.0
    # should this be added in a way so the user can check how the sequence looks like ?
    def run_sim(self, sim_time):
        """Run the simulation

        Args:
            sim_time (float): Time[s] to simulate starting from 0
        """
        self.device_setup = lo.DeviceSetup.from_dict(
            data=self.descriptor,
            server_host="localhost",
            server_port=SERVER_PORT,
            setup_name=self.name,
        )
        # create a session
        self.sim_session = lo.Session(self.device_setup)
        # connect to session
        self.sim_device = self.sim_session.connect(do_emulation=True)
        self.exp = self.sim_session.compile(self.experiment, compiler_settings=COMPILER_SETTINGS)

        # Plot simulated output signals with helper function
        plot_simulation(
            self.exp,
            start_time=0,
            length=sim_time,
            plot_width=10,
            plot_height=3,
        )



class Port:
    """Abstract interface for instrument parameters.

    These parameters are exposed to the user through :class:`qibolab.channels.Channel`.

    Drivers should subclass this interface and implement the getters
    and setters for all the parameters that are available for the
    corresponding instruments.

    Each port is identified by the ``name`` attribute.
    Note that the type of the identifier can be different of each port implementation.
    """

    name: str
    """Name of the port that acts as its identifier."""
    offset: float
    """DC offset that is applied to this port."""
    lo_frequency: float
    """Local oscillator frequency for the given port.
    Relevant only for controllers with internal local oscillators.
    """
    lo_power: float
    """Local oscillator power for the given port.
    Relevant only for controllers with internal local oscillators.
    """
    # TODO: Maybe gain, attenuation and power range can be unified to a single attribute
    gain: float
    """Gain that is applied to this port."""
    attenuation: float
    """Attenuation that is applied to this port."""
    power_range: int
    """Similar to attenuation (negative) and gain (positive) for (Zurich instruments)."""
    filters: dict
    """Filters to be applied to the channel to reduce the distortions when sending flux pulses.

    Useful for two-qubit gates.
    Quantum Machines (:class:`qibolab.instruments.qm.QMOPX`) associate filters to channels
    but this may not be the case in other instruments.
    """
